#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

# MASS-LITE object manager program: follows the output data files of ESO-DIMM to
# trace the target change and DIMM measurements beginning.

# Idea: to use the intranet FTP of the current DIMM log-file 
# for comparison of the last line content with the previous one.
#  - If the star coordinates change, the new star (HR) name is extracted from
# the comparison with the DIMM target list. This name is set as a new object
# requested by SV as GET OBJECT request. OBJECT=0 is reserved for unrecognized
# object number cases.
#  - If the star is the same and the data line changes, this means that the
# DIMM measurements are under way and the MASS measurement may be performed.
# This state is reflected by the flag DIMMRUN=1 requested as GET DIMMRUN
#  - If the data line does not change withing $dimm_period, this means that 
# DIMM measurements are stopped and flag switches to DIMMRUN=0.
#
# Apart from DIMMRUN value, the additional data are provided by GET-command, 
# namely: SEEING, AIRMASS, UTC, OBJECT
# 
# The script should be started in the VLT environment (i.e. where the FTP access 
# specified below is possible) and is supposed to be
# accessed by the Supervisor program which manipulates the Turbina MASS program.
# The INIT command starts the permanent background FTPing of the DIMM archive,
# The PARK command stops it. The RUN and STOP commands are not implemented.
# The GET command retrieves the DIMM parameters and running flag DIMMRUN.
# If GET is started in PARKED state, it will cause the component to retrieve
# the parameters only once. It is an alternative mode of use (retrieve on 
# request).
#
# The functionality is implemented using the FTP access to the current DIMM 
# log-file with a given periodicity (ftp_period).

#---------------------- PROGRAM SETTINGS --------------------
# Port number:
set port 6240

# Identification:
set ident "getdimm.tcl v0.01"

# Star catalogue file name for getting HR number from (DIMM) coordinates:
# Consists of a pair of values: RA_DEC HR (E.g. "15846_-613411 591")
set star_cat dimmstar.lst

# FTP program flags:
#SH set ftp_opt "-nu"

# FTP server address:
#SH set ftp_address "wlsasm"
set ftp_address "ftp://anonymous:kolja@dragon/pub/log/dimm.log"

# FTP commands for getting data:
#SH set ftp_cmd "user anonymous pclmass@eso.org\ncd pub/log\nget dimm.log\nquit"

# Periodicity of the requests to DIMM log-file via FTP, in MILLISECONDS:
set ftp_period 10000

# Duration of "freshness" of the DIMM file: if the last line doesn't change
# over this time, DIMM is considered to be non-measuring. This time, in SECONDS,
# must be more than the SUM of $ftp_period and the longest possible DIMM 
# accumulation time.
set dimm_period [expr 5*60+$ftp_period/1000]

# DIMM log-file structure: put here the field names and field numbers (starting 
# with 0!) of all 
# parameters which are interesting to be extracted and accessible by GET :
array set dimm_log { utc 0 ra 3 dec 4 seeing 39 zendist 5 vmag 6 airmass 7 }

# MASS-files directory (trailing slash is essential):
set massfile_dir /usr/share/turbina/out/

# FTP area for public access to MASS-files:
set ftp_dir /var/ftp/pub/masslog/

#--------------------- END OF SETTINGS -----------------------------

# Commands interface
source compcmd.tcl

# Accessing the retrieved and processed data from Supervisor:
proc cmd_get pars {
	# global array of parameter values cmp_pvar(param)=value and comp.status:
	global cmp_pvar cmp_status ftp_after
	
	# here we allow to retrieve all global variables:
	foreach p $pars {
		if {[info exists ::$p]} {
			upvar #0 $p par
			set cmp_pvar($p) $par 
		} elseif {$p=="putmass"} {
			set cmp_pvar($p) [putmass]
		} else {
			set cmp_status ERSYN; break
		}
	}
}

# After-identifier of the ftp procedure:
# "0" means absence of scheduled start and, if getdimm{} finds it =0, it must
# not restart itself.
set ftp_after 0

# Array of stars: index name is a pair of coordinates RA_DEC, value is HR number:
set catalogue(0) 0

proc cmd_init {} {
	global cmp_status ftp_after star_cat catalogue object
	if {$cmp_status=="PARKED"} {
		# Read the star catalogue:
		if {[catch {set fid [open $star_cat]} errMsg]} {
			puts "Catalogue open error: $errMsg"
			set cmp_status ERFAT
			return
		} else {
			# read catalogue:
			set num 0
			while {[gets $fid l]>=0} {
				incr num
				set catalogue([lindex $l 0]) [lindex $l 1]
			}
			close $fid
		}
		# Here we simply start the ftp activity on the background:
		puts "INITIALIZED : $num stars"
	}
	# Allow the requests chaining:
	set ftp_after 1
	# start requesting (in a chain):
	getdimm
	set cmp_status READY
}

proc cmd_park {} {
	global cmp_status ftp_after
	# Cancel the next scheduled start of the procedure:
	after cancel $ftp_after
	set ftp_after 0
	set cmp_status PARKED
	puts "PARKED"
}

# content of the previous line of the DIMM log:
set prevline ""

# Star coordinates of the previous line
set prevcrd ""

# Time of getting the last fresh line, in sec-1969:
set dimm_sec 0

# Flag of the FTP error output:
set ftp_error 0

# Putting the mass-file in the FTP area: called upon arrival GET PUTMASS
proc putmass {{outname mass.log}} {
	global massfile_dir ftp_dir
	# Put the data in the public FTP area:
	if {[catch {\
		set massfile [exec ls $massfile_dir | grep -e 0 | grep mass | tail -n 1]
		exec cp $massfile_dir$massfile $ftp_dir/$outname
	} errMsg]} {
		puts "Cannot put massfile to FTP area: $errMsg"
		return 0
	} else {
		return 1
	}
}

proc getdimm {} {
	global ftp_after ftp_period ftp_opt ftp_address ftp_cmd cmp_status
	global dimm_sec dimm_log prevline prevcrd dimm_period dimmrun catalogue
	global ftp_error object

	# Access DIMM log-file:
#SH	set err [catch {\
#		set res [exec echo $ftp_cmd | ftp $ftp_opt $ftp_address] \
#			} errMsg]
	set err [catch {\
		set res [exec ftp $ftp_address] \
			} errMsg]
	# Check that there was no bad output or ftp running problem:
	if {$err } {
		puts "FTP command failed: $errMsg"
		set cmp_status ERFAT
		return
	} elseif {$res!=""} {
		# most probably - no dimm.log file: output once:
		if {$ftp_error==0} {
			puts "FTP error: $res"
			set ftp_error 1
		}
		foreach par [concat [array names dimm_log] object dimmrun] {
			global $par
			set $par 0
		}
	} else {
	# reset flag of the output of the error of FTP:
	set ftp_error 0
	set lastline [exec tail -n 1 dimm.log]
	set firstline [exec head -n 1 dimm.log]
	# check whether it's only a header:
	if {$lastline==$firstline} {
		puts "Only a header in a file..."
		foreach par [concat [array names dimm_log] object dimmrun] {
			global $par
			set $par 0
		}
	} else {

	if {$lastline!=$prevline} {	
		# new line added: extract parameters:
		puts ""
		foreach par [array names dimm_log] {
			global $par
			set $par [set p [lindex $lastline $dimm_log($par)]]
			puts -nonewline "$par = $p "
		}
		puts ""
		
		# See whether the object has changed:
		set crd "$ra\_$dec"
		if {$crd!=$prevcrd} {
			# Cross-reference coordinates with HR:
			set hr "-"
			if {[catch {set hr $catalogue($crd)}] || $hr=="-"} {
				puts "Star crd $crd not found!"
				set object 0
			} else {
				set object $hr
				puts "Star changed: coordinates $crd, HR=$hr"
			}
			set prevcrd $crd
		}
		
		set dimm_sec [clock seconds]
		set prevline $lastline
		# check that this line is not old: case of restart of getdimm
		set dsec [clock scan [string map {T \ } $utc] -gmt true]
		puts "DIMM sec = $dsec, clock = [clock seconds]"
		set dimmrun [expr ([clock seconds] - $dsec)<$dimm_period]
	} else {
		# No fresh line: check how old the previous data are:
		if {[expr ([clock seconds] - $dimm_sec)>$dimm_period] && \
			$dimmrun} { 
			puts "DIMM stopped?"
			set dimmrun 0 
		}
		if {$dimmrun} {
			#put a hash-mark of ongoing measurements:
			puts -nonewline "="
			flush stdout
		}
	}
	} 
	# - this brace closes the if-body for no FTP error (err=0)
	} 
	# - this closes the firstline=lastline comparison
	# restart procedure after a specified delay:
	if {$ftp_after!="0"} {
		set ftp_after [after $ftp_period getdimm]
	}  else {puts "Stop"}
}

open_server $port
exit
