#!/bin/sh
# the next line restarts using tcl/tk interpreter\
exec wish "$0" "$@"

#
#  SUPERVISOR: Observation process manager. (C) Shatsky, SAI, RUSSIA (2003-04)
# e-mail: kolja@sai.msu.ru
#
namespace eval sv {
# version:
variable version 0.22

# The only hardcoded setting is a name of configuration file:
variable cfg_file_name [pwd]/sv.cfg

# List of real components (not aliases) which need INIT,PARK etc:
variable comp_list [list]

#List of aliases:
variable alias_list [list]

# SV parameteres with defaults obligatory for SV functionality (?=no default):
array set SV_check_list {tmout 10 oscen ? cscen ? revive_time 0 \
	emergency_sys "echo \"SV termination!\"" interactive 0 start_monitor 1}

# Component parameters obligatory for functionality (NO status,chanid in CFG!):
array set comp_check_list { host 127.0.0.1 port ? ident ? optional 0 \
	status "UNDEFINED" chanid "closed"}

# 
# Reading the SV configuration file (SV and components parameters)
#
# @param file Configuration file name (see its syntax in sv.cfg preamble)
# @return	number of parameters set
#
proc read_cfg file {

	if {[catch {set in [open $file r]}]} {
		#ENOCFG:
		problem "No cfg $file found" ENOCFG
	}
	# Until first "component xxx" line the parameters are those of SV:
	set comp SV
	# Declare SV as a global (array) variable:
	global SV
	# Create the SV-procedure for accessing SV-parameters from scenaria:
	set s "{ return \$::SV\(\$param\) }"
	uplevel #0 proc SV param $s
	set linenum 0
	set parnum 0
	while {[gets $in line] >= 0} {
		incr linenum
		# Locate possible comment character:
		if {[set ci [string first \# $line]] >=0} { 
			# remove righthand comment after ci:
			set line [string range $line 0 [expr $ci-1]]
		}
		# delete possible leading/trailing spaces:
		set line [string trim $line]	
		# locate parameter name end:
		set dp [string wordend $line 0]
		set sl [string length $line]
		if {$dp<$sl} {
			set pname [string tolower [string range $line 0 [expr $dp-1]]]
			set pvalue [string trim [string range $line \
				[string wordstart $line	$dp] end]]
			if {$pname=="component"} {
				# declaration of the next component section
				set comp $pvalue
				# new component or addition to previously defined?:
				if {[lsearch $sv::comp_list $comp]<0} {
					# add new component to the list of known components
					lappend sv::comp_list $pvalue
				}
				# declare the global array of component parameters:
				global $pvalue
				upvar #0 $comp cmp
				# create the same name procedure for accessing array
				set s "{ return \$::$pvalue\(\$param\) }"
				uplevel #0 proc $pvalue param $s
			} elseif {$pname=="alias"} {
				# declaration that the current component is an alias of another
				# delete this component if it is different from what it should
				# alias:
				if {[info exists ::$comp] && \
					[array get $comp *]!=[array get	$pvalue *]} {
					# Copy all elements into target of aliasing:
					array set $pvalue [array get $comp *]
					add_log "CFG: ATTENTION! $comp is alias of $pvalue but has own parameters"
					add_log "CFG: Move $comp el-s [array get $comp *] to $pvalue!"
					unset ::$comp
				}
				# delete this component from the list as well, since no 
				# personal init/stop/park is needed for it:
				set i [lsearch $sv::comp_list $comp]
				set sv::comp_list [lreplace $sv::comp_list $i $i]
				upvar #0 ::$pvalue ::$comp

				if {[lsearch $sv::alias_list $comp]<0} {
					# add new alias comp. to the list of known aliases:
					lappend sv::alias_list $comp
				}
				
			} else {
				upvar #0 $comp cmp
				set cmp($pname) $pvalue
				incr parnum
			}
		} else {
			#EBADCFG:
			if {$dp>0} {
				problem "Single word in cfg $file, not a comment (line $linenum)"\
				 EBADCFG
			}
		}
	}
	close $in
	
	# Check CFG:
	foreach p [array names sv::SV_check_list] {
		if {![info exists SV($p)] && \
			[set SV($p) $sv::SV_check_list($p)]=="?"} {
				problem "CFG file: no SV($p) found" ENOPCFG
		}
	}
	
	foreach c $sv::comp_list {
		upvar #0 $c cmp
		set cmp(name) $c			
		foreach p [array names sv::comp_check_list] {
			if {![info exists cmp($p)] && \
				[set cmp($p) $sv::comp_check_list($p)]=="?"} {
				problem "CFG file: no $c\($p\) found" ENOPCFG				
			}
		}
	}
	
	return $parnum
}

#-------------------------- COMMAND SERVICE --------------------------------
# Array of lists of "after" results and commands, for which these
# after-timers were set, indexed by cmd_id,comp pairs: e.g. 
# cmd_array(0,OBJM) = {after#1 {GET ID}}
# After receiving a reply, the element must be unset.
variable cmd_array

# Common sv-namespace variable - command identifier
variable cmd_id 0

# For displaying the commands history tail:
variable answered_cmd_list [list]
variable answered_cmd_list_n 4

#
# Sending a command to a component.
#
# @param comp component name (as in CFG)
# @param kwd  command keyword (GET, RUN, etc)
# @param args >=0 switches or parameter=value pairs
# [@param] "&" as last argument: return after 1st reply even if it's OK WAIT=...
# @return command identifier on success, -1 on error
#

proc cmd {comp kwd args} {
	global $comp SV
	upvar #0 $comp cmp
	if {[info globals $comp]==""} {
		problem "A command to non-existent component $comp" ECMPNEX
		return -1
	}
		
	# Send a command string to the component and register and schedule its 
	# timeout:
	set chanid $cmp(chanid)
	# - chanid here is assigned in connect{}

	# Detect request for background execution command: no waiting for completion, 
	# only for the first reply:
	if {[set i [lsearch $args "\&"]]>=0} {
		set args [lreplace $args $i $i]
		set bkgrexec 1
	} else {
		set bkgrexec 0
	}

	# Pause-button braking: except for cmd_issue(),stop_park()-originated calls
	if {$sv::paused} {
		set level [info level]
		if {[info level]==1 || \
			[set caller [lindex [info level -1] 0]]!="sv::cmd_issue" && \
			$caller!="sv::stop_park"} {	vwait sv::paused }
	}

	set kwd [string toupper $kwd]

	set cmd "$kwd"
	# remove braces:
	foreach a $args { set cmd "$cmd $a" }
	if {$chanid=="closed" || 
		[catch {puts $chanid "$sv::cmd_id $cmd"; flush $chanid}]} {
		#ECMDDSC:
		problem "Command \"$cmd\" to disconnected $cmp(name)" ECMDDSC \
			$cmp(name) $cmd
		return -1
	} else {
		#ECMDLOS: Register the sent command in an array
		set s [subst {{Timeout waiting for $cmp(name) reply} ECMDLOS $cmp(name) {$cmd}}]
		set sv::cmd_array($sv::cmd_id,$cmp(name)) [list \
			[after $::SV(tmout) sv::problem $s] $cmd]
	}
	add_log "->$comp $sv::cmd_id $cmd"
	set prev_cmd_id $sv::cmd_id
	# Modify the command identifier according to the 0..65535,0.. cycling rule:
	if {$sv::cmd_id==65535} { set sv::cmd_id 0 } else { incr sv::cmd_id }

	# Delay the return from the command until it's replied:
	# Wait until variable is unset (if no & supplied). WAIT-reply sets it back!
	while {[info exists sv::cmd_array($prev_cmd_id,$cmp(name))]} {
			vwait sv::cmd_array($prev_cmd_id,$cmp(name)) 
			if {$bkgrexec} break
	}
	return $prev_cmd_id
}

#
# Reply whether the command with a given identifier is under execution
#
# Searchs the array of commands for occurrence of the given command
# identifier.
# @param comid the command identifier (as returned by cmd{})
# @return 1 - the command is still under execution, 0 - command is over 
# (or was not not issued at all)
#
proc is_cmd comid {
	return [expr [llength [array names sv::cmd_array $comid,*]]>0]
}

#
# Wait until one of the command(s) execution is finished
#
# Each time the command list is updated (by cmd{}, get_reply{}), checks whether 
# one of the commands (which identifiers are provided) is over (removed
# by get_reply{} from the command list) and if so - return the identifier of the
# finished command. If several commands are over at the time of checking, the
# identifier which stands earlier in the list of arguments is returned.

# @param args one or several command identifiers
# @return command identifier (one of those in args) which command is over
#
proc wait_cmd args {
#	puts "Enter in WAIT_CMD $args"
	while 1 {
		foreach c $args {
			if {![is_cmd $c]} {
#				puts "Return from WAIT_CMD $args"
				return $c
			}
		}
		vwait sv::cmd_array
	}
}

#
# Wait for some time in seconds to pass and return.
#
# Each 1/2sec the update is called, to unfreeze SV application and process
# pending events. Use this procedure for integer-second delays only.
# 
# @param intrp interpreter name (supplied automatically in alias-making)
# @param sec waiting time in seconds, INTEGER
proc wait_sec {intrp sec {addlog true}} {
#	puts "WAIT_SEC for $sec from $intrp"
	if {$addlog} {add_log "WAIT $sec seconds..."}
	set till [expr [clock seconds] + $sec]
	while {[clock seconds]<$till && [interp exists $intrp]} { 
		after 500; update; after 500; update
	}
#	puts "RETURN fom WAIT_SEC for $sec"
}

# Here we include the utilities for parsing the command/reply strings and
# composing the reply/commands from list of parameter[=value]s:
source parsecmd.tcl

#
# Common handler for all component channels reading which parses their replies
#
# This handler is called from the individual component channel handlers - the
# procedures get_reply_<componentname> created in connect{} and bound to the
# channel by the fileevent command.
# @param comp component name
# @return not defined
#
proc get_reply comp {
	global $comp
	upvar #0 $comp cmp
	if {[gets $cmp(chanid) reply] < 0} {
		#ECMPDSC: end of file condition: disactivate current handler
		problem "Lost connection to $comp" ECMPDSC $cmp(name)
		return
	}
	add_log "<-$comp $reply"
	
	# Create reply-list from the reply-string:
	set rl [parse_command $reply]

	# Extract command identifier and keyword:
	set ci [ lindex $rl 0 ]
	set kwd [lindex $rl 1]
	# Remove COMID and KWD from reply leaving the parameter=value's only:
	set rl [lreplace $rl 0 1]
#puts "receive1: cmd_array($comp)=[array get sv::cmd_array *,$comp]"
#puts "After=[lindex $sv::cmd_array($ci,$comp) 0]"
	# Cancel the timeout timer for this command and delete it from list:
	if {[catch {set cmd [lindex $sv::cmd_array($ci,$cmp(name)) 1]}]} {
		#ECMDID: if caugth then no timer for such COMID was registered for COMP
		problem "Unexpected reply (no command with ci=$ci sent to $comp)"\
		 ECMDID $cmp(name)
	}
	after cancel [lindex $sv::cmd_array($ci,$cmp(name)) 0]
	
	# Parse rest of reply and assign the replied quantities to the component 
	# parameters:
	
	if {[lsearch $rl switch]>=0} {
		#ECMDPAR:
		problem "Parsing reply from $comp: param. without value in $reply"\
		 ECMDPAR $cmp(name)
	} else {
		# remove wait-parameter for the case if it was a delayed result:
		array unset $cmp(name) wait
		# set the replied parameter(s) with value(s) for the component:
		array set $cmp(name) $rl
		if {[llength [array names $cmp(name) wait]]>0} {
			# Postpone final reply getting a bit more:
			set wms [expr int($cmp(wait) * 1000 + $::SV(tmout))]
#puts "Postpone reply timeout by another $wms ms"
			#ECMDLOW: Register another timer of the command in an array
			set s [subst {{Timeout waiting for $comp reply (after wait)}\
			 ECMDLOW $cmp(name) {$cmd}}]
			set sv::cmd_array($ci,$cmp(name)) [list 	[after $wms sv::problem $s]\
				[lindex $sv::cmd_array($ci,$cmp(name)) 1] ]
#puts "postpone: cmd_array($comp)=[array get sv::cmd_array *,$comp]"
		}
	}
	# Remove the command from the array if there was no wms set (WAIT parsed).
	# We could not do it before parsing! since there must be no period of
	# respective element absence (vwait in cmd{}).
	if {[info exists wms]==0} {
		# GUI: append the current answered command to the list and remove oldest
		lappend sv::answered_cmd_list "$ci ->$comp $cmd  <- $reply"
		if {[llength $sv::answered_cmd_list]>$sv::answered_cmd_list_n} {
			set sv::answered_cmd_list [lreplace $sv::answered_cmd_list 0 0]
		}
		array unset sv::cmd_array $ci,$cmp(name)
	}
	
	switch $kwd {
		"ERROR"
			{ 	
				problem "Command cannot be executed: $cmp(status)"\
				 ECMPSTA $cmp(name) $cmd
			}
		"OK"
			{ }
		default
			{
				problem "Unknown keyword: $reply from $comp" ECMDKWD $cmp(name)
			}
	}

#puts "receive: cmd_array($comp)=[array get sv::cmd_array *,$comp]" 
#puts "$comp params:[array get ::$comp]"
}


#------------------------- TIME SERVICE ------------------------------------
#
# Return the local time stamp according to the adopted format
#
# Time stamp is adopted in format YYYY-MM-DD HH-MM-SS and composed from the
# machine UT time to be consistent with other program log files.
# @param sec1969	Time is seconds after 1969 (Unix integer time). Current by
# 	default
# @return Time stamp string
#
proc time_stamp {{sec1969 -1}} {
	if {$sec1969==-1} {
		set sec1969 [clock seconds]
	}
	return [clock format $sec1969 -format "%Y-%m-%d %H:%M:%S" -gmt true ]
}

#------------------------ ESTABLISH COMMUNICATION ---------------------------

#
# Connect component(s) listed in arguments
#
# Tries to create the socket-connection to components using their host and port
# parameters listed in CFG. Channel is configured with default parameters. 
# If connected, creates the channel reading handler and binds the channel to it.
# If status of some components is not PARKED - park them after all are 
# connected.
 
# The reading handler simply calls get_reply{}.
# Then checking of the component identifier is made: error is caused if no match
# between reply and cfg name is found. Finally, component status is asked.
# @param complist component identifier(s)
#
proc connect complist {
	# comp. list for parking
	set forparking [list]
	foreach comp $complist {
		global $comp 
		upvar #0 $comp cmp
		add_log "Connect $comp at $cmp(host) port $cmp(port)"
		# Check if comp is started:
		
		if {[catch {set t [socket $cmp(host) $cmp(port)]} ]} {
			#ENOCMP:
			problem "Cannot connect to $comp" ENOCMP $comp
        } else {
			set cmp(chanid) $t
			# Create the procedure for handling the replies (Overwrites existing):
			uplevel #0 "proc get_reply_$comp {} { sv::get_reply $comp }"
			# Set it as a component channel reading handler:
			fileevent $cmp(chanid) readable get_reply_$comp
        }
		
		set ident_cfg $cmp(ident)
		set c1 [cmd $comp GET IDENT]
		if {![string match -nocase $cmp(ident) $ident_cfg]} {
			problem "$comp ident $cmp(ident) does not match $ident_cfg from cfg"\
			 ENMCMP $comp
		}
		cmd $comp GET STATUS
		if {$cmp(status)!="PARKED"} {lappend forparking $comp}
	}
	# Park if not parked?
	if {[llength $forparking]} {stop_park $forparking}
}

#
# Disconnect component(s) closing its channel
#
# Channel identifier is assigned "closed" for recognition by cmd{}.
# @param complist component identifier(s) list
# @return 1 if channel was open, 0 if channel was already closed
#
proc disconnect complist {
	set res 0
	foreach comp $complist { 
		global $comp
		upvar #0 $comp cmp
		# the channel reading handler will be unset automatially now:
		if {$cmp(chanid)!="closed"} {
			set res 1
			add_log "Disconnect $comp"
			# Channel may appear not found if already closes: catch this error
			catch {close $cmp(chanid)}
			set cmp(chanid) closed
			set cmp(status) UNKNOWN
			# delete all pending commands to the component (to cause 
			# possible wait_cmd{} to return):
			array unset sv::cmd_array *,$comp
		}
	}
	return $res
}

#------------------------ INITIALIZATION/PARKING ---------------------------
#
# Initialize component(s)
#
# Issue INIT command to all components listed in complist and wait until all
# components finish initialization.
# @param complist name of the list with component names or "all" for all comps.
#
proc initialize complist {
	# Issue the Initialize command to all components:
	foreach comp $complist {
		lappend initcmdlist [cmd $comp INIT &]
	}
	# wait for all init commands are over:
	add_log "Wait for components to finish initialization..."
	foreach c $initcmdlist { wait_cmd $c }
	add_log "Initialization is over."	
}

#
# Stop and park components
#
# Issue STOP NOW and PARK commands and wait until all component finish parking.
# @param complist name of the list with component names.
# Note: if component status is PARKED, the commands are not sent.
#
proc stop_park complist {
	add_log "Stop and park components $complist"
	set parkcmdlist [list]
	foreach comp $complist {
		upvar #0 $comp cmp
		if {$cmp(chanid)=="closed" || $cmp(status)=="PARKED"} {
			add_log "$comp: status is $cmp(status) - Skipped."
			continue
		}
		# here component might stuck, in principle. cmd..& + is_cmd() is needed,
		# but Turbina, for example, WAITs basetime seconds before sending
		# OK status=ready. Delayed checking to be implemented?
		cmd $comp STOP NOW
		if {$cmp(status)=="BUSY"} {
			#ECMPNST: 
			problem "$comp is busy after STOP NOW" ECMPNST $comp
		}
		if {[is_cmd [set p [cmd $comp PARK &]]]} {lappend parkcmdlist $p}
	}
	# wait for all park commands are over:
	if {[llength $parkcmdlist]} {
		add_log "Wait for components to finish parking..."
		foreach c $parkcmdlist { wait_cmd $c }
		add_log "Parking finished."
	}

}

#------------------------ SCENARIA MANAGEMENT -------------------------------
#
# Check the scenaria (for observations, monitoring)
#
# The checking of scenaria existence is performed. Currently that's all.
# @param args list of scenaria file names
# @return 1 if all scenaria are Ok, 0 otherwise
#
proc check_scen args {
	foreach s $args { 
		if {![file readable $s]} {
			problem "Scenario $s is not readable" EBADSCE
			return
		}
	}
	add_log "Scenaria $args are Ok"
}

#
# Start the scenario in the interpreter
#
# Create a safe interpreter (for network safety reasons etc.) and create 
# aliases of common and additional procedures. Then source scenario with the 
# created interpreter on the background (after 10ms delay).
# NOTE that interaction of this procedure and stop_obs/monitor{} is a tricky 
# place in SV programming!
# @param interp_var_name Interpreter variable name to be assigned with created
#                        interpreter identifier
# @param scen scenario file name
# @param add_cmd_list list of additional commands to be aliased ({} for empty)
# @param start_flag_name name of global variable to be set to 1 upon starting
#                        scenario.
#
#

proc start_scen {interp_var_name scen add_cmd_list {start_flag_name ""}} {
	global SV errorCode
	upvar $interp_var_name intr
	set intr [interp create -safe]

	# creating commands in scenaria:
	foreach c [concat {cmd wait_cmd is_cmd add_log stop_park initialize} \
		SV $sv::comp_list $sv::alias_list $add_cmd_list] { $intr alias $c $c}
	# waiting-commands may be dangerous: we'll stop waiting if caller is deleted:
	foreach c {wait_sec} {$intr alias $c $c $intr}
	
	set start_command "add_log \"Start scenario: source $scen\""
	if {$start_flag_name!=""} {
		set start_command "$start_command; set $start_flag_name 1"
	}
	# below catch catches the eval of deleted interp. commands when scenario is
	# stopped:
	set start_command [subst -nocommands "$start_command;
		if {\[catch {$intr invokehidden source $scen} sceErr\]} "]
	set start_command [subst -nocommands -novariables "$start_command {\
			if {\[lindex \$errorCode 0\]!=\"CORE\" ||\
			   \[lindex \$errorCode 1\]!=\"IDELETE\"} {\
			   	sv::problem \$sceErr ECMDSCE $intr\
			}\
		};
		add_log \"$scen sourcing is canceled\""]
		
	after 10 $start_command
}

#
# Evaluate a command in the observations interpreter, catch-protected
# @param intr interpreter name 
# @param args command with optional parameters
# @return result of command evaluation in interpreter
#
proc cmd_scen {intr args} {
	if {[catch {set res [$intr eval $args]} sceErr]} {
		 sv::problem $sceErr ECMDSCE $intr
	} else { return $res }
}

# dummy command for alias in a "deleted" scenario interpreter:
proc null_cmd args {
	error "Deleted interpreter command" {} {CORE IDELETE}
	return -1
}

# List of interpreters of scenaria which are under stopping:
variable stop_ips [list]
#
# Stop scenario
#
# Before deletion of scenario interpreter, calls end{} if found in scenario. Then
# deletes the safe interpreter for the scenario created in start_scen{}
# thus stopping scenario evaluation (next command in scenario fails with an error
# "attempt to call a command in deleted interpreter" which is caught in a script
# started from start_scen{} on a background. NOTE that this is a FINE POINT in
# such a programming of SV!)
# @param interp_var_name name of interpreter variable
# @return success of stopping
#
proc stop_scen interp_var_name {
	upvar $interp_var_name intr
	if {$intr==0 || [lsearch $sv::stop_ips $intr]>=0} {return 0}
	lappend sv::stop_ips $intr
	if {[$intr eval info procs end]!=""} {
		sv::cmd_scen $intr end
	}
	# disactivate the commands : "deleted" interpreter might live 
	# for a while: cmd add_log stop_park initialize wait_sec...
	foreach c [$intr aliases] { $intr alias $c sv::null_cmd }
	interp delete $intr
	add_log "Interpreter $intr deleted."
	set sv::stop_ips [lreplace $sv::stop_ips \
		[set i [lsearch $sv::stop_ips $intr]] $i]
	set intr 0
	return 1
}

#================== OBSERVATIONS AND CIRCUMSTANCE SCENARIA MANAGEMENET =======
# Flag of observations going on. Precisely, this means that safe Tcl interpreter
# is created and executes the observations scenario (see start_obs{}).
variable is_observations_now 0

#
# Return flag of "observations now" for calling scenario.
#
# @return $is_observations_now
#
proc is_observations_now {} {return $sv::is_observations_now}

# Tcl interpreter identifier for running observation scenario
variable tclobs 0
# flag of protection against (monitor-originated) starting of observations
variable disable_start_obs 0

#
# Start observations scenario
#
# Call the scenario starter if allowed to start obs-s. Then, create the
# obs. scenario buttons window. For the time of starting obs-s, disable
# starting. 
#
proc start_obs {} {
	global SV
	# detect whether the stop_obs was forced before with disable_start_obs==1
	if {$sv::disable_start_obs || $sv::tclobs!=0} return
	# avoid starting observations during initialization (i.e. no-reenter):
	set sv::disable_start_obs 1
	
	sv::start_scen sv::tclobs $SV(oscen) {} sv::is_observations_now
	# Create the scenario buttons if specified:
	if {[array names SV button*]!=""} {
		set w [toplevel .obsbut]
		wm title $w "Observations scenario control"
		wm protocol $w WM_DELETE_WINDOW {}
		set f [frame $w.buttons]
		foreach but [array names SV button*] {
			set bname [lindex $SV($but) 0]
			set baction [lreplace $SV($but) 0 0]
			pack [button $f.$but -text $bname -command \
				"sv::cmd_scen $sv::tclobs $baction"] -side left -padx 10
		}
		pack $f -pady 1
	}
	set sv::disable_start_obs 0
	add_log "Observations scenario started with $sv::tclobs"
}

#
# Stop observations 
#
# @param disable_start flag of protection of immediate restart of
#        observations, 0 by default. Used for manual stopping and
#        termination of SV. Otherwise monitor may restart obs-s again.
#
proc stop_obs {{disable_start 0}} {
	global SV
	if {![interp exists $sv::tclobs]} return
	if {$disable_start} {
		set sv::disable_start_obs $disable_start
	}
	if {![stop_scen sv::tclobs]} return
	# remove the scenario buttons window if created
	if {[array names SV button*]!=""} { destroy .obsbut }
	set sv::is_observations_now 0
	add_log "Observations stopped"
}

# Tcl interpreter identifier for running the monitor scenario
variable tclmon 0
#
# Start circumstance monitor scenario
#
# Calls scenario starter if allowed, with additional commands for aliasing. 
#
proc start_monitor {} {
	global SV
	if {$sv::tclmon!=0} return
	sv::start_scen sv::tclmon $SV(cscen) "start_obs stop_obs is_observations_now"
	add_log "Circumstance monitor scenario started with $sv::tclmon"
}

#
# Stop circumstance monitor:
#
proc stop_monitor {} {
	if {![interp exists $sv::tclmon] || ![stop_scen sv::tclmon]} return
	sv::add_log "Circumstance monitor stopped"	
}

# SV paused status flag:
variable paused 0

#-------------------------- INTERFACE, LOGGING ------------------------------
# Log file descriptor
variable logfileid 0
#
# Open a log-file of the session as YYMMDDsv.log, where date is UTC date
#
proc open_log {} {
	set sv::logfileid [open [clock format [clock seconds] \
		-format "%y%m%dsv.log" -gmt true ] a]
	add_log "Supervisor V$sv::version session open"
}
#
# Close log-file
#
proc close_log {} {
	if {$sv::logfileid!=0} {close $sv::logfileid; set sv::logfileid 0 }
}
#
# Add a log line in the sv.log and on a console with a time stamp.
#
proc add_log line {
	puts "[time_stamp] $line"
	if {$sv::logfileid!=0} {
		puts $sv::logfileid "[time_stamp] $line"
		flush $sv::logfileid
	}
}

#
# Execute the user-defined command to inform the external world about error.
# In case SV runs interactively (SV(interactive)==1), do nothing.
# Called from problem{}
#
proc call_emergency {} {
	global SV
	if {!$SV(interactive) && [catch {exec $SV(emergency_sys) &}]} {
		add_log "CANNOT EXECUTE $sv::cfg_file_name->emergency_sys!" 
	}
}
#
# Handle errors 
#
# Procedure takes some actions for correction the problem in situ and returns
# the control back to calling procedure - if it is possible; or cancels the 
# normal SV work with help terminate{} and call_emergency{} procedures.
#
# @param message Problem error message
# @param code error code for automatic deciphering
# Optional arguments which are supplied and used depending on the error
# and action which may be performed to handle situation:
# @param comp	Component name
# @param cmd	Command sent
#
variable problem_list [list]
proc problem {message code {comp NO} {cmd NO}} {
	global SV argv errorInfo

	if {$sv::terminating} {
		add_log "ERROR: $message while terminating: IGNORED"
		return
	}

	# avoid infinite recursion of the procedure with the same code and component:
	# (we cannot do this via "info level" command since we use cmd{} to handle
	# the situation and get_reply_COMP is invoked on the top level.
	set curpbm [list $code $comp $cmd]
	if {[lsearch $sv::problem_list $curpbm]>=0} {
		# Current problem call with $curpbm found. Return.
		return
	} else {
		# Problem $curpbm encounters for the first time
		lappend sv::problem_list $curpbm
		# This is output for tracing the problems resolution only:
#		add_log "Problem list=$sv::problem_list"
	}
	add_log "ERROR: $message"
	upvar #0 $comp cmp
	
	switch -glob $code {
		ECMDDSC {
			# Wait a bit to slower trafic down if it is a chain of commands:
			after 1000;
		}

		E*CFG - EBADSCE {	
			# Startup errors: simply stop SV
			exit
		}
		
		EN*CMP {
			terminate
			exit
		}
		
		ECMDSCE - ECMPNEX {
			# Scenario errors: $comp comtains scenario interpreter ID.
			# Check that it is still existing interpreter (not deleted):
			if {![interp exists $comp]} return
			add_log $errorInfo
			if { $SV(interactive) } {
				switch [tk_dialog .int_err "$comp scenario error" \
				"$message (see log)" "" 2 \
				"Stop without end{}" "Call end{} if exists, stop" "Skip"] {
					0 {
						# Overwrite end{} if exists with empty:
						$comp eval proc end args {return}
						if {$comp==$sv::tclobs} stop_obs else stop_monitor
					}
					1 {
						if {$comp==$sv::tclobs} stop_obs else stop_monitor
					}
					2 {}
				}
			} else {
				call_emergency
				terminate
			}
		}
		
		ECMDID - ECMPNST - ECMDPAR - ECMDKWD -		
		ECMDLO* - ECMPDSC - ECMPFAT {
			# Component problem:
			# if connection is lost - first close channel to prevent its channel
			# from continuous calling of event handler:
			if {[string match ECMPDSC $code]} {
				set comp_disconnected [disconnect $comp]
			}
			set handled 0
			# Try to call the error handler in either scenario script:
			foreach interp [list $sv::tclobs $sv::tclmon] {
				if { $interp!="0" && [$interp eval info procs error_handler]!=""} {
					set handled [cmd_scen $interp error_handler $code $comp]
					# If handled - make switch below doing skip:
					if {$handled} {set handled 1; break}
				}
			}
			if {!$handled && $SV(interactive)} {
				set default_action "Terminate"
				if {$cmp(optional)} {set default_action "Close $comp"}
				set handled [tk_dialog .int_err "Component error" \
				"$comp: $message" "" 1 $default_action "Skip"]
			}
			if {!$handled} {
				if {[string match ECMDLO* $code] || \
					[string match ECMPDSC $code] || \
					[string match ECMPNST $code]} {
					# do nothing - no parking is possible
				} else {
					# otherwise - park bad component before doing other things:
					sv::stop_park $comp
				}
				if {$code!="ECMPDSC"} {set comp_disconnected [disconnect $comp]}
			
				if {$comp_disconnected} {
					call_emergency
				
					if {$cmp(optional)} {
						add_log "Component $comp is optional, so we closed it"
					} else {
						# Component is essential for normal work: try to revive:
						terminate
						if {$SV(revive_time)>0} {
							add_log "Shell will sleep for $SV(revive_time) sec and then SV restarts"
							exec echo "/bin/sleep $SV(revive_time)\n./sv.tcl $argv &\n\
							rm sv_again" >sv_again
							exec chmod u+x sv_again
							exec ./sv_again &
						}
						exit
					}
				}
				# otherwise - this is secondary error already...
			}
		}
		
		ECMPSTA {
			switch $cmp(status) {
			
				"BUSY" {
					add_log "wait while $comp is $cmp(status)..."
					vwait ::${comp}(status)
					add_log "Ok: $comp is $cmp(status): send $cmd to $comp again"
					cmd $comp $cmd
				} 
			
				"ERANG" - "ERSYN" {
					# Was the error caused by a dialog to this component?
					if {[info exists ::combo_box_$comp] && [string match \
						-nocase $cmd [set ::combo_box_$comp]]} {
						add_log "Error from Dialog with $comp: IGNORED"
					} else {
						sv::problem $message ECMPFAT $comp $cmd
					}
				}
			
				default {
#					add_log "IGNORED"
					sv::problem $message ECMPFAT $comp $cmd
				}
			}
		}
		
		default {
			# Here we are if we missed some (new?) error code:
			call_emergency
			error $message "Unexpected problem code"
		}
	}

	# On return: remove the handled problem from the list:
	set i [lsearch $sv::problem_list $curpbm]
	set sv::problem_list [lreplace $sv::problem_list $i $i]
}

# Tix package loaded flag
variable istix 0
#
# Create the GUI for monitoring the SV and components state. Supports
# configuration from SV cfg file ("show"-lists of variables for monitoring);
# component names are the buttons which allow to issue the command to the given
# component.
#
proc create_window {} {
	set sv::istix [expr ![catch {package require Tix}]]

	wm withdraw .
	set w [toplevel .t]
	wm title .t "Supervisor v$sv::version"
	wm protocol .t WM_DELETE_WINDOW sv::terminate
	
	# create frame-table with sub-frames-columns each for its component items
	set f [frame $w.ft]
	pack [label $f.tlabel -text Components] -side top
	pack $f -fill x -padx 2 -pady 0
	
	set f [frame $w.fcmp -borderwidth 3]
	foreach c $sv::comp_list { 
		global $c
		upvar #0 $c cmp
		set fc [frame $f.f$c -bd 3 -width 100 -relief groove]
		# button with component name:
		menubutton $fc.cn -text $c -state normal -menu $fc.cn.menu
		set m [menu $fc.cn.menu]
		$m add command -label Command -command "sv::cmd_dialog $c"
		label $fc.status -textvariable ${c}(status) -fg red
		pack $fc.cn $fc.status -fill x -padx 2 -pady 0 -side top
		if {[info exists cmp(show)]} {
			foreach v $cmp(show) {
			if {$v=="data"||$v=="status"} continue
				frame $fc.v$v
				pack [label $fc.v$v.lab -text "$v:" ] -side left
				pack [label $fc.v$v.val -textvariable ${c}($v) -fg blue] -side \
					left
				pack $fc.v$v -anchor w
			}
		}
		$m add command -label Disconnect -command "sv::disconnect $c"
		$m add command -label Connect -command "sv::connect $c"
		pack $fc -side left -anchor n -expand true -fill y
	}
	pack $f -pady 0 -expand true
	
	# Create a section with components data lines:
	set f [frame $w.fdata]
	foreach c $sv::comp_list {
		global $c
		upvar #0 $c cmp
		if {[info exists cmp(show)]==0 || [lsearch $cmp(show) data]<0} continue
		set fc [frame $f.f$c]
		pack [label $fc.dlabel$c -text "$c DATA:"] -side left -anchor w
		pack [label $fc.data$c -textvariable ${c}(data) -fg blue] -side left -anchor w
		pack $fc -padx 2 -pady 1 -anchor w
	}
	pack $f -pady 2 -anchor w

	# Create the message with observations/monitoring state:
	set f [frame $w.obs]
	pack [label $f.lab -text "Activity:" ] -side left
	pack [label $f.val -text "undefined" -fg blue] -side left
	pack $f -pady 2
	
	
	# Create the window with the last and delayed-execution commands:
	set f [frame $w.cmd -takefocus 0]
	pack [label $f.label -text "Last commands:"]
	pack [listbox $f.cmdlist -takefocus 0 -selectmode none -height \
		[expr [llength $sv::comp_list]+$sv::answered_cmd_list_n]  -width 60]
	pack $f -padx 2 -pady 2 -anchor w
	trace variable sv::cmd_array wu sv::update_cmd_array
	
	# Buttons to manipulate observations state:
	set f [frame $w.cmdobs]
	pack [button $f.stop -text "Stop Observations" -state disabled \
		-command {sv::stop_obs 1}] -side left -padx 10
	pack [button $f.start -text "Start Observations" \
		-command {set sv::disable_start_obs 0; sv::start_obs}]  -side left -padx 10
	pack [button $f.enable -text "Allow Observations" \
		-command {set sv::disable_start_obs 0}]  -side left -padx 10
	pack $f -pady 1
	trace variable sv::is_observations_now w sv::update_buttons
	trace variable sv::disable_start_obs w sv::update_buttons
	trace variable sv::tclobs w sv::update_buttons
	trace variable sv::stop_ips w sv::update_buttons

	# Button to pause/resume circ. monitor and observations evaluation:
	set f [frame $w.pausecmd]
	pack [button $f.pause -text "Pause" -state disabled \
		-command {set sv::paused [expr !$sv::paused]; \
		if {$sv::paused} {add_log "PAUSE"}}]
	pack $f -pady 1
	trace variable sv::paused w sv::update_buttons

	# Buttons to manipulate circumstance monitor:
	set f [frame $w.cmdmon]
	pack [button $f.stop -text "Stop monitor" -state disabled \
		-command {sv::stop_monitor}] -side left -padx 10
	pack [button $f.start -text "Start monitor" \
		-command {sv::start_monitor}]  -side left -padx 10
	pack [button $f.terminate -text "Terminate" \
		-command sv::terminate]  -side left -padx 10
	pack $f -pady 1
	trace variable sv::tclmon w sv::update_buttons
}

#
# Depending on $is_observations_now, activate or disable buttons in GUI
#
# This procedure serves to reflect the modifications of is_observations_now
# variable in the SV window (see "man n trace"). Also, the "preparation" stage
# is detected when tclobs interpreter is already created while is_observations_now
# is still 0.
# Starting with SV version 0.18 this procedure embeds also the functions of 
# Pause and Start/Stop monitor buttons state changing.
#
# @param name1 accessed variable name
# @param name2 element name if name1 is an array
# @param op access operation code (w - write, r - read, u - unset; w normally)
#
proc update_buttons {name1 name2 op} {
	if {$sv::tclobs!=0} {
		if {[lsearch $sv::stop_ips $sv::tclobs]<0} {
			.t.obs.val configure -text "Observations"
		} else {
			.t.obs.val configure -text "Cancelling observations..."
		}
	} else {
		if {$sv::tclmon==0} {
			.t.obs.val configure -text "Doing nothing"
		} elseif {[lsearch $sv::stop_ips $sv::tclmon]<0} {
			.t.obs.val configure -text "Monitoring conditions"
		} else {
			.t.obs.val configure -text "Cancelling monitoring..."
		}
	}

	if {$sv::disable_start_obs && $sv::tclobs==0} {
		.t.cmdobs.enable configure -state active
	} else {
		.t.cmdobs.enable configure -state disabled
	}

	foreach sce {obs mon} {
		upvar sv::tcl$sce intr
		if {$intr==0} {
			.t.cmd$sce.stop configure -state disabled
			.t.cmd$sce.start configure -state active
		} elseif {[lsearch $sv::stop_ips $intr]<0} {
			.t.cmd$sce.stop configure -state active
			.t.cmd$sce.start configure -state disabled
		} else {
			.t.cmd$sce.stop configure -state disabled
			.t.cmd$sce.start configure -state disabled
		}
	}

	# pause/resume button: active if either interpreter alive:
	if {$sv::tclobs!=0 || $sv::tclmon!=0} {
		.t.pausecmd.pause configure -state active
		if {$sv::paused} {
			foreach b {cmdobs.stop cmdobs.start cmdobs.enable cmdmon.stop
				cmdmon.start} {
				.t.$b configure -state disabled
			}
			.t.pausecmd.pause configure -text "Resume"
			set i [lsearch -glob [.t.obs.val configure] "-text *"]
			.t.obs.val configure -text \
				"[lindex [lindex [.t.obs.val configure] $i] 4] paused"
		} else {
			.t.pausecmd.pause configure -text "Pause"
		}
	} else {
		.t.pausecmd.pause configure -state disabled
	}
}

#
# Operator's dialog with a component from Tk window:
# @param comp Component name
# Dialog is implemented in two varieties - with a single Tix widget tixComboBox
# and, for the case of no Tix available, with usual Tk widgets.
#
proc cmd_dialog comp {
	# detect whether the dialog with a given component already exists:
	if {[info exists dia_label_$comp]} return
	set t [toplevel .dia$comp]
	wm title $t "SV command dialog with $comp"
	wm protocol $t WM_DELETE_WINDOW \
		"destroy $t; unset ::dia_label_$comp"
	# create two global variables with a label and ComboBox ID (first call only):
	set ::dia_label_$comp "Enter command to $comp"
	if {![info exists ::combo_box_$comp]} {set ::combo_box_$comp {}}
	if {$sv::istix} {
		tixComboBox $t.cmd -selectmode browse -editable true -dropdown true\
		-padx 20 -pady 4 -variable combo_box_$comp\
		-labelside top -fancy true -history true -historylimit 20 \
		-command "sv::cmd_issue $comp"
		$t.cmd subwidget label configure -width 100 -textvariable ::dia_label_$comp
		foreach cmnd { INIT RUN STOP PARK FREE GET SET } { $t.cmd insert end $cmnd }
		pack $t.cmd
	} else {
		# no Tix loaded:
		pack [label $t.label -textvariable ::dia_label_$comp]
		pack [entry $t.cmd -textvariable ::combo_box_$comp] -side left
		pack [button $t.send -text "Send command" \
		-command "sv::cmd_issue $comp \$::combo_box_$comp"]  -side left -padx 10
	}
	
}

#
# Send a command to the component and put a reply in the sv::label_cont var.
# @param comp Component name
# @param cmd command keyword and arguments (if needed)
#
# Note: The text with a reply is written in the global variable
# ::dia_label_$comp which is created in cmd_dialog{}
proc cmd_issue {comp cmd} {
	cmd $comp $cmd
	set ::dia_label_$comp \
		"Transaction: [lindex $sv::answered_cmd_list end]\n New command:"
}

#
# Update the view of the commands list in the SV window: invoked by trace var.
# First creates the new list of delayed commands, then deletes the old
# displayed list and adds the sorted concatenation of the delayed commands and
# answered commands
#
proc update_cmd_array {name1 name2 op} {
	set cl [list]
	foreach n [array names sv::cmd_array] {
		set s "$n [lindex $sv::cmd_array($n) 1]"
		set s [string replace $s [set i [string first , $s]] $i " ->"]
		lappend cl $s
	}
	set cls [lsort -dictionary [concat $cl $sv::answered_cmd_list]]
	.t.cmd.cmdlist delete 0 [expr [.t.cmd.cmdlist size] - 1]
	foreach e $cls {
		.t.cmd.cmdlist insert end $e}
}

#
# Terminate SV
#
# Stops monitor interpreter (if it still executes something directly), stop
# observations (if under way) and do all the job to stop SV correctly.
# Note that the procedure may be called from different places, but has a sence
# only once. So, to avoid reentering from the one hand and immediate return from
# the other, the vwait-return block is used.
#
variable terminating 0
proc terminate {} {
	if {$sv::terminating} {vwait sv::terminating; return}
	set sv::terminating 1
	add_log "TERMINATE"
	stop_monitor
	stop_obs 1
	# This ensures that all components are parked. Reentering of stop_park{}
	# is not dangerous but saves when observations are under stopping to the
	# moment of terminate{} call:
	sv::stop_park $sv::comp_list
	disconnect $sv::comp_list
	# release possible vwait-ing in pause_resume{}
	if {$sv::paused} {set sv::paused 0}
	close_log
	destroy .t
	set sv::terminating 0
}

###################### Starting supervisor:
#
# Starting the normal algorithm of SV functionality
#
proc main {} {
	open_log
	read_cfg $sv::cfg_file_name
	create_window
	global SV
	check_scen $SV(cscen) $SV(oscen)
	connect $sv::comp_list
	# arrival of TERM signal will remove the SV window causing tkwait...t to end
	# this demands TclX package to be loaded:
	if {[catch {package require Tclx}]} {
		tk_messageBox -type ok -parent .t -icon warning -message \
		"To enable remote correct SV kill-termination, install TclX package"
	} else {
		signal trap SIGTERM { sv::terminate } 
	}
	if {$SV(start_monitor)} start_monitor
	tkwait window .t
}

namespace export -clear cmd is_cmd wait_cmd wait_sec start_obs \
  stop_obs add_log is_observations_now stop_park initialize

}

# import those commands of namespace sv which are used in scenaria:
namespace import sv::cmd sv::is_cmd sv::wait_cmd sv::wait_sec sv::start_obs \
  sv::stop_obs sv::add_log sv::is_observations_now sv::stop_park sv::initialize

sv::main
exit
